"""
Script to generate pyi stubs by patching and calling mypy's stubgen tool.

There are two entry-points which are designed to call this script:
 - `make pystubs` is called during local development to generate the
   stubs and copy them into the git repo to be committed and reviewed.
 - in CI, the cibuildwheel action is used to validate that the stubs match what
   has been committed to the repo.

The depdendencies for the script are defined in pyproject.toml.
"""

from __future__ import absolute_import, annotations, division, print_function

import mypy.stubgen
import mypy.stubgenc
from mypy.stubgenc import SignatureGenerator, DocstringSignatureGenerator

from stubgenlib.siggen import (
    AdvancedSignatureGenerator,
    AdvancedSigMatcher,
)
from stubgenlib.utils import add_positional_only_args


PY_TO_STDVECTOR_ARG = "float | typing.Iterable[float]"


class OIIOSignatureGenerator(AdvancedSignatureGenerator):
    sig_matcher = AdvancedSigMatcher(
        signature_overrides={
            # signatures for these special methods include many inaccurate overloads
            "*.__ne__": "(self, other: object) -> bool",
            "*.__eq__": "(self, other: object) -> bool",
        },
        arg_type_overrides={
            # FIXME: Buffer may in fact be more accurate here
            ("*", "*", "Buffer"): "numpy.ndarray",
            # these use py_to_stdvector util
            ("*.ImageBufAlgo.*", "min", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "max", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "black", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "white", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "sthresh", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "scontrast", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "white_balance", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "values", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "top", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "bottom", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "topleft", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "topright", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "bottomleft", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "bottomright", "object"): PY_TO_STDVECTOR_ARG,
            ("*.ImageBufAlgo.*", "color", "object"): PY_TO_STDVECTOR_ARG,
            # BASETYPE & str are implicitly converible to TypeDesc
            ("*", "*", "*.TypeDesc"): "Union[TypeDesc, BASETYPE, str]",
            # list is not strictly required
            ("*.ImageOutput.open", "specs", "list[ImageSpec]"): "typing.Iterable[ImageSpec]",
        },
        result_type_overrides={
            # FIXME: is there a way to use std::optional for these?
            ("*.ImageOutput.create", "object"): "ImageOutput | None",
            ("*.ImageOutput.open", "object"): "ImageOutput | None",
            ("*.ImageInput.create", "object"): "ImageInput | None",
            ("*.ImageInput.open", "object"): "ImageInput | None",

            # ("*.TextureSystem.imagespec", "object"): "ImageSpec | None",

            # if you return an uninitialized unique_ptr to pybind11 it will convert to `None` (see #4685)
            ("*.ImageInput.read_native_deep_*", "DeepData"): "DeepData | None",

            # pybind11 has numpy support, so it may be possible to get it to emit these types
            # by using py::numpy in our wrapper code.
            ("*.ImageInput.read_*", "object"): "numpy.ndarray | None",
            ("*", "Buffer"): "numpy.ndarray",
            ("*.get_pixels", "object"): "numpy.ndarray | None",

            # For results, `object` is too restrictive (produces spurious errors during type analysis)
            ("*.getattribute", "object"): "typing.Any",
            ("*.ImageSpec.get", "object"): "typing.Any",

            ("*.ImageBufAlgo.histogram", "*"): "tuple[int, ...]",
            # pybind11 treats std:vector<T> as list[T], but we want tuple[T, ...]
            # our custom code to convert vector to tuple obscures the contained type.
            ("*.ImageBufAlgo.isConstantColor", "*"): "tuple[float, ...] | None",
            ("*.ImageBufAlgo.color_range_check", "*"): "tuple[int, ...] | None",

            ("*.TextureSystem.imagespec", "object"): "ImageSpec | None",
            ("*.TextureSystem.texture", "tuple"): "tuple[float, ...]",
            ("*.TextureSystem.texture3d", "tuple"): "tuple[float, ...]",
            ("*.TextureSystem.environment", "tuple"): "tuple[float, ...]",

            ("*.ImageBuf.getpixel", "tuple"): "tuple[float, ...]",
            ("*.ImageBuf.interppixel*", "tuple"): "tuple[float, ...]",
            ("*.ImageSpec.get_channelformats", "tuple"): "tuple[TypeDesc, ...]",
        },

        property_type_overrides={
            # FIXME: this isn't working
            ("*.ParamValue.value", "object"): "typing.Any",
        },
    )

    def process_sig(
        self, ctx: mypy.stubgen.FunctionContext, sig: mypy.stubgen.FunctionSig
    ) -> mypy.stubgen.FunctionSig:
        # Analyze the signature and add a '/' argument if necessary to mark
        # arguments which cannot be access by name.
        return add_positional_only_args(ctx, super().process_sig(ctx, sig))


class InspectionStubGenerator(mypy.stubgenc.InspectionStubGenerator):
    def get_sig_generators(self) -> list[SignatureGenerator]:
        return [
            OIIOSignatureGenerator(
                fallback_sig_gen=DocstringSignatureGenerator(),
            )
        ]


mypy.stubgen.InspectionStubGenerator = InspectionStubGenerator  # type: ignore[attr-defined,misc]
mypy.stubgenc.InspectionStubGenerator = InspectionStubGenerator  # type: ignore[misc]


def get_colored_diff(old_text: str, new_text: str):
    """
    Generates a colored diff between two strings.

    Returns:
        A string containing the colored diff output.
    """
    import difflib

    red = '\033[31m'
    green = '\033[32m'
    reset = '\033[0m'

    diff = difflib.unified_diff(
        old_text.splitlines(keepends=True),
        new_text.splitlines(keepends=True),
        lineterm="",
    )
    lines = []
    for line in diff:
        if line.startswith('-'):
            lines.append(f"{red}{line}{reset}")
        elif line.startswith('+'):
            lines.append(f"{green}{line}{reset}")
        else:
            lines.append(line)
    return "".join(lines)


def main() -> None:
    import argparse
    import pathlib
    import os
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--out-path",
        default="out",
        help="Directory to write the stubs."
    )
    parser.add_argument(
        "--validate-path",
        default=None,
        help="If provided, compare the generated stub to this file. Exits with code 2 if the "
             "contents differ."
    )
    args = parser.parse_args()
    out_path = pathlib.Path(args.out_path)
    print(f"Stub output directory: {out_path}")

    # perform import so we can see the traceback if it fails.
    import OpenImageIO

    sys.argv[1:] = ["-p", "OpenImageIO", "-o", str(out_path)]
    mypy.stubgen.main()
    source_path = out_path.joinpath("OpenImageIO", "OpenImageIO.pyi")
    if not source_path.exists():
        print("Stub generation failed")
        sys.exit(1)

    dest_path = out_path.joinpath("OpenImageIO", "__init__.pyi")
    print(f"Renaming to {dest_path}")
    os.rename(source_path, dest_path)

    new_text = dest_path.read_text()
    new_text = (
        "#\n# This file is auto-generated. DO NOT MODIFY!  "
        "Run `make pystubs` to regenerate\n#\n\n"
    ) + new_text
    dest_path.write_text(new_text)

    if args.validate_path and os.environ.get("GITHUB_ACTIONS", "false").lower() == "true":
        # in CI, validate that what has been committed to the repo is what we expect.
        validate_path = pathlib.Path(args.validate_path)

        print("Validating stubs against repository")
        print(f"Comparing {dest_path} to {validate_path}")

        old_text = validate_path.read_text()

        if old_text != new_text:
            print("Stub verification failed!")
            print("Changes to the source code have resulted in a change to the stubs.")
            print(get_colored_diff(old_text, new_text))
            print("Run `make pystubs` locally and commit the results for review.")
            print("The resulting __init__.pyi file will be uploaded as an artifact on this job.")
            sys.exit(2)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import sys
import os
import shlex
import subprocess


# -- Project information -----------------------------------------------------

# The master toctree document.
master_doc = 'index'

default_role = 'code'

project = 'OpenImageIO'
copyright = '2008-present, Contributors to OpenImageIO'
author = 'Larry Gritz'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# LG addition: we search for it in the CMakeLists.txt so we don't need to
# keep modifying this file:
version = '0.0'
release = '0.0.0'
import re
version_regex = re.compile(r'set \(OpenImageIO_VERSION \"?((\d+\.\d+)\.\d+)\.\d+\"?\)')
f = open('../../CMakeLists.txt')
for l in f:
    aa=re.search(version_regex, l)
    if aa is not None:
       release = aa.group(1)
       version = aa.group(2)
       break
f.close()
print ("OIIO docs version = {}, release = {}".format(version, release))



# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

# add custom extensions directory to python path
sys.path.insert(0, os.path.join(os.path.abspath(os.path.dirname(__file__)), 'extensions/sphinxtr'))
sys.path.insert(0, os.path.join(os.path.abspath(os.path.dirname(__file__)), '../../ext/breathe'))
#import html_mods
#import latex_mods

extensions = [
              'breathe',
              'sphinx_tabs.tabs'
 ]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'
#html_theme = 'astropy-sphinx-theme'
# html_theme = 'sphinx_rtd_theme'
html_theme = 'furo'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# Breathe options
breathe_projects = { "oiio": "../../build/doxygen/xml" }
breathe_default_project = "oiio"
breathe_domain_by_extension = {'h': 'cpp'}
breathe_default_members = ()
primary_domain = 'cpp'
highlight_language = 'cpp'


read_the_docs_build = os.getenv('READTHEDOCS') == 'True'

if read_the_docs_build:
    print ("cwd =", os.getcwd())
    print ("checkpoint -- rtd build")
    if not os.path.exists('../../build/doxygen') :
        os.makedirs ('../../build/doxygen')
    print ("checkpoint 2 -- rtd build")
    subprocess.call('echo "Calling Doxygen"', shell=True)
    subprocess.call(['doxygen'], shell=True)
    subprocess.call('echo "Ran Doxygen"', shell=True)



# Copyright Contributors to the OpenImageIO project.
# SPDX-License-Identifier: Apache-2.0
# https://github.com/AcademySoftwareFoundation/OpenImageIO

import os
import sys
import platform
import subprocess

_here = os.path.abspath(os.path.dirname(__file__))

# Set $OpenImageIO_ROOT if not already set *before* importing the OpenImageIO module.
if not os.getenv("OpenImageIO_ROOT"):
    if all([os.path.exists(os.path.join(_here, i)) for i in ["share", "bin", "lib"]]): 
        os.environ["OpenImageIO_ROOT"] = _here

if platform.system() == "Windows":
    # Python wheel module is dynamically linked to the OIIO DLL present in the bin folder.
    _bin_dir = os.path.join(_here, "bin")
    if os.path.exists(_bin_dir):
        os.add_dll_directory(_bin_dir) 
    elif sys.version_info >= (3, 8):
        # This works around the python 3.8 change to stop loading DLLs from PATH on Windows.
        # We reproduce the old behavior by manually tokenizing PATH, checking that the 
        # directories exist and are not ".", then add them to the DLL load path.
        # This behavior is disabled by default, but can be enabled by setting the environment
        # variable "OPENIMAGEIO_PYTHON_LOAD_DLLS_FROM_PATH" to "1"
        if os.getenv("OPENIMAGEIO_PYTHON_LOAD_DLLS_FROM_PATH", "0") == "1":
            for path in os.getenv("PATH", "").split(os.pathsep):
                if os.path.exists(path) and path != ".":
                    os.add_dll_directory(path)

from .OpenImageIO import * # type: ignore # noqa: F401, F403, E402

__doc__ = """
OpenImageIO is a toolset for reading, writing, and manipulating image files of 
any image file format relevant to VFX / animation via a format-agnostic API 
with a feature set, scalability, and robustness needed for feature film 
production.
"""[1:-1]


def _call_program(name, args):
    bin_dir = os.path.join(os.path.dirname(__file__), 'bin')
    return subprocess.call([os.path.join(bin_dir, name)] + args)

def _command_line():
    name = os.path.basename(sys.argv[0])
    raise SystemExit(_call_program(name, sys.argv[1:]))



#!/usr/bin/env python3
# Copyright 2018 Nico Weber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Converts one (or several) .ninja_log files into chrome's about:tracing format.

If clang -ftime-trace .json files are found adjacent to generated files they
are embedded.

Usage:
    ninja -C $BUILDDIR
    ninjatracing $BUILDDIR/.ninja_log > trace.json
"""

import json
import os
import optparse
import re
import sys


class Target:
    """Represents a single line read for a .ninja_log file. Start and end times
    are milliseconds."""
    def __init__(self, start, end):
        self.start = int(start)
        self.end = int(end)
        self.targets = []


def read_targets(log, show_all):
    """Reads all targets from .ninja_log file |log_file|, sorted by start
    time"""
    header = log.readline()
    m = re.search(r'^# ninja log v(\d+)\n$', header)
    assert m, "unrecognized ninja log version %r" % header
    version = int(m.group(1))
    assert 5 <= version <= 6, "unsupported ninja log version %d" % version
    if version == 6:
        # Skip header line
        next(log)

    targets = {}
    last_end_seen = 0
    for line in log:
        if line.startswith('#'):
            continue
        start, end, _, name, cmdhash = line.strip().split('\t') # Ignore restat.
        if not show_all and int(end) < last_end_seen:
            # An earlier time stamp means that this step is the first in a new
            # build, possibly an incremental build. Throw away the previous data
            # so that this new build will be displayed independently.
            targets = {}
        last_end_seen = int(end)
        targets.setdefault(cmdhash, Target(start, end)).targets.append(name)
    return sorted(targets.values(), key=lambda job: job.end, reverse=True)


class Threads:
    """Tries to reconstruct the parallelism from a .ninja_log"""
    def __init__(self):
        self.workers = []  # Maps thread id to time that thread is occupied for.

    def alloc(self, target):
        """Places target in an available thread, or adds a new thread."""
        for worker in range(len(self.workers)):
            if self.workers[worker] >= target.end:
                self.workers[worker] = target.start
                return worker
        self.workers.append(target.start)
        return len(self.workers) - 1


def read_events(trace, options):
    """Reads all events from time-trace json file |trace|."""
    trace_data = json.load(trace)

    def include_event(event, options):
        """Only include events if they are complete events, are longer than
        granularity, and are not totals."""
        return ((event['ph'] == 'X') and
                (event['dur'] >= options['granularity']) and
                (not event['name'].startswith('Total')))

    return [x for x in trace_data['traceEvents'] if include_event(x, options)]


def trace_to_dicts(target, trace, options, pid, tid):
    """Read a file-like object |trace| containing -ftime-trace data and yields
    about:tracing dict per eligible event in that log."""
    ninja_time = (target.end - target.start) * 1000
    for event in read_events(trace, options):
        # Check if any event duration is greater than the duration from ninja.
        if event['dur'] > ninja_time:
            print("Inconsistent timing found (clang time > ninja time). Please"
                  " ensure that timings are from consistent builds.")
            sys.exit(1)

        # Set tid and pid from ninja log.
        event['pid'] = pid
        event['tid'] = tid

        # Offset trace time stamp by ninja start time.
        event['ts'] += (target.start * 1000)

        yield event


def embed_time_trace(ninja_log_dir, target, pid, tid, options):
    """Produce time trace output for the specified ninja target. Expects
    time-trace file to be in .json file named based on .o file."""
    for t in target.targets:
        o_path = os.path.join(ninja_log_dir, t)
        json_trace_path = os.path.splitext(o_path)[0] + '.json'
        try:
            with open(json_trace_path, 'r') as trace:
                for time_trace_event in trace_to_dicts(target, trace, options,
                                                       pid, tid):
                    yield time_trace_event
        except IOError:
            pass


def log_to_dicts(log, pid, options):
    """Reads a file-like object |log| containing a .ninja_log, and yields one
    about:tracing dict per command found in the log."""
    threads = Threads()
    for target in read_targets(log, options['showall']):
        tid = threads.alloc(target)

        yield {
            'name': '%0s' % ', '.join(target.targets), 'cat': 'targets',
            'ph': 'X', 'ts': (target.start * 1000),
            'dur': ((target.end - target.start) * 1000),
            'pid': pid, 'tid': tid, 'args': {},
            }
        if options.get('embed_time_trace', False):
            # Add time-trace information into the ninja trace.
            try:
                ninja_log_dir = os.path.dirname(log.name)
            except AttributeError:
                continue
            for time_trace in embed_time_trace(ninja_log_dir, target, pid,
                                               tid, options):
                yield time_trace


def main(argv):
    usage = __doc__
    parser = optparse.OptionParser(usage)
    parser.add_option('-a', '--showall', action='store_true', dest='showall',
                      default=False,
                      help='report on last build step for all outputs. Default '
                      'is to report just on the last (possibly incremental) '
                      'build')
    parser.add_option('-g', '--granularity', type='int', default=50000,
                      dest='granularity',
                      help='minimum length time-trace event to embed in '
                      'microseconds. Default: %default')
    parser.add_option('-e', '--embed-time-trace', action='store_true',
                      default=False, dest='embed_time_trace',
                      help='embed clang -ftime-trace json file found adjacent '
                      'to a target file')
    (options, args) = parser.parse_args()

    if len(args) == 0:
      print('Must specify at least one .ninja_log file')
      parser.print_help()
      return 1

    entries = []
    for pid, log_file in enumerate(args):
        with open(log_file, 'r') as log:
            entries += list(log_to_dicts(log, pid, vars(options)))
    json.dump(entries, sys.stdout)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



"""
Script that is called when buildling the stubs during local development, via
`make pystubs` or `cmake --target pystubs`.
"""
import platform

import argparse
import subprocess
import sys


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--repo-root")
    parser.add_argument("--output-dir")
    parser.add_argument("--python-version")
    parser.add_argument("--cibuildwheel-version")

    args = parser.parse_args()

    version_major, version_minor = args.python_version.split(".")

    # detect the platform so that we can avoid running docker under emulation.
    if platform.system() == "Darwin":
        # On Mac, `uname -m`,  $CMAKE_HOST_SYSTEM_PROCESSOR, and platform.machine() returns
        # x86_64 even on arm64 processors (i.e. Apple Silicon) if cmake or python is running under
        # emulation (Rosetta). sysctl.proc_translated tells us if we're under emulation.
        arch = subprocess.check_output(["uname", "-m"], text=True).strip()
        if arch == "x86_64":
            process_uses_emulation = subprocess.check_output(
                ["sysctl", "-in", "sysctl.proc_translated"], text=True
            ).strip()
            if process_uses_emulation == "1":
                arch = "arm64"
    else:
        arch = platform.machine()
    if arch == "arm64":
        arch = "aarch64"
    python_build_id = f"cp{version_major}{version_minor}-manylinux_{arch}"
    print(f"Building {python_build_id}")

    try:
        subprocess.check_call(
            [
                "uv",
                "tool",
                "run",
                f"--python={sys.executable}",
                f"cibuildwheel@{args.cibuildwheel_version}",
                f"--output-dir={args.output_dir}",
                f"--only={python_build_id}",
                ".",
            ],
            cwd=args.repo_root,
        )
    except FileNotFoundError:
        print("\nERROR: You must install uv to build the stubs. See https://docs.astral.sh/uv/\n")
        sys.exit(1)


if __name__ == "__main__":
    main()
